@@ -1,3 +1,58 @@
+'''Service using multiprocessing to perform CPU-bound operations
+making use of multiple CPUs (i.e. cores) on the machine.
+
+Each "worker" runs in a separate process. There can be multiple
+workers forming a sequence or an ensemble. Exact CPU allocation---which
+CPUs are used by which workers---can be controlled
+to achieve high efficiency.
+
+The "interface" to the service resides in the "main process".
+Two usage patterns are supported, namely (concurrent) individual
+calls to the service to get individual results, or flowing
+a (potentially unlimited) stream of data through the service
+and consuming the outcoming stream of results. Each usage
+supports a async API and an async API.
+
+The user's main work is implementing the worker operations.
+These are conventional sync functions. Another work of trial and error
+is experimenting with CPU allocations among workers to achieve
+best performance (throughput and latency).
+
+In contrast, the interface and scheduling code in the main process
+is all provided, and usually does not need much customization.
+
+Typical usage:
+
+    class MyServlet1(Servlet):
+        def __init_(self, ..):
+            ...
+
+        def call(self, x):
+            ...
+
+    class MyServlet2(Servlet):
+        def __init__(self, ...):
+            ...
+
+        def call(self, x):
+            ...
+
+    class MyServer(SequentialServer):
+        def __init__(self, ...):
+            ...
+            self.add_servlet(MyServlet1, ...)
+            self.add_servlet(MyServlet2, ...)
+
+    with MyServer(...) as server:
+        y = server.call(x)
+
+        output = server.stream(data)
+        for y in output:
+            ...
+
+Reference: [Service Batching from Scratch, Again](https://zpz.github.io/blog/batched-service-redesign/).
+This article describes roughly version 0.7.2.
+'''
 
 import asyncio
 import concurrent.futures
@@ -35,7 +90,7 @@ class TotalTimeout(TimeoutError):
 
 class Servlet(metaclass=ABCMeta):
     # Typically a subclass needs to enhance
-    # `__init__` and implement `__call__`.
+    # `__init__` and implement `call`.
 
     @classmethod
     def run(cls, *,
@@ -61,7 +116,7 @@ class Servlet(metaclass=ABCMeta):
                  batch_size: int = None,
                  batch_wait_time: float = None):
         '''
-        `batch_size`: max batch size; see `__call__`.
+        `batch_size`: max batch size; see `call`.
 
         `batch_wait_time`: seconds, may be 0; the total time span
             to wait for one batch after the first item has arrived.
@@ -81,7 +136,7 @@ class Servlet(metaclass=ABCMeta):
         benchmarks. So beware of this in benchmarking.
 
         Remember to pass in `batch_size` in accordance with the implementation
-        of `__call__`.
+        of `call`.
         '''
         self.batch_size = batch_size or 0
         self.batch_wait_time = batch_wait_time or 0
@@ -93,9 +148,9 @@ class Servlet(metaclass=ABCMeta):
             uid, x = q_in.get()
             try:
                 if batch_size:
-                    y = self.__call__([x])[0]
+                    y = self.call([x])[0]
                 else:
-                    y = self.__call__(x)
+                    y = self.call(x)
                 q_out.put((uid, y))
 
             except Exception as e:
@@ -163,7 +218,7 @@ class Servlet(metaclass=ABCMeta):
                             n_batches, batch_size_max, batch_size_min, batch_size_mean)
 
             try:
-                results = self.__call__(batch)
+                results = self.call(batch)
             except Exception as e:
                 err = MPError(e)
                 for uid in uids:
@@ -182,7 +237,7 @@ class Servlet(metaclass=ABCMeta):
             self._start_single(q_in=q_in, q_out=q_out, q_err=q_err)
 
     @abstractmethod
-    def __call__(self, x):
+    def call(self, x):
         # If `self.batch_size == 0`, then `x` is a single
         # element, and this method returns result for `x`.
         #
@@ -212,7 +267,7 @@ class Servlet(metaclass=ABCMeta):
         raise NotImplementedError
 
 
-class _Server(metaclass=ABCMeta):
+class MPServer(metaclass=ABCMeta):
     MP_CLASS = mp
     # This class attribute is provided because in some cases
     # one may want to use `torch.multiprocessing`, which is
@@ -329,6 +384,14 @@ class _Server(metaclass=ABCMeta):
              enqueue_timeout: Union[int, float] = None,
              total_timeout: Union[int, float] = None,
              ):
+        '''
+        Sometimes a subclass may want to override this method
+        to add preprocessing of the input and postprocessing of
+        the output, while calling `super().call(...)` in the middle.
+
+        However, be careful to maintain consistency between the
+        `call`/`async_call` methods and the `stream`/`async_stream` methods.
+        '''
         enqueue_timeout, total_timeout = self._resolve_timeout(
             enqueue_timeout=enqueue_timeout, total_timeout=total_timeout,
         )
@@ -550,7 +613,7 @@ class _Server(metaclass=ABCMeta):
         while not q_err.empty():
             uid, err = q_err.get_nowait()
             fut = futures.pop(uid, None)
-            if fut is None:  # timed-out in `__call__`
+            if fut is None:  # timed-out in `call` or `async_call`
                 logger.info(
                     'got error for an already-cancelled task: %r', err)
                 continue
@@ -684,7 +747,7 @@ class _Server(metaclass=ABCMeta):
                         raise e
 
 
-class SequentialServer(_Server):
+class SequentialServer(MPServer):
     '''
     A sequence of operations performed in order,
     the previous op's result becoming the subsequent
@@ -735,11 +798,7 @@ class SequentialServer(_Server):
         return self._q_in_out[:1]
 
 
-Server = SequentialServer
-# For back compat. Will remove.
-
-
-class EnsembleServer(_Server):
+class EnsembleServer(MPServer):
     '''
     A number of operations performed on the same input
     in parallel, the list of results gathered and combined
@@ -841,7 +900,7 @@ class SimpleServer(SequentialServer):
                 super().__init__(batch_size=batch_size)
                 self._kwargs = kwargs
 
-            def __call__(self, x):
+            def call(self, x):
                 return func(x, **self._kwargs)
 
         self.add_servlet(SimpleServlet,
